import os
import sys
# os.chdir("../..")
sys.path.insert(0, os.getcwd())

import pytest
import numpy as np
import pandas as pd
import dreamtools as dt

def test_DataFrame_with_series():
  dt.REFERENCE_DATABASE = dt.Gdx("../test.gdx")
  s = dt.Gdx("../test.gdx")
  dt.time(2025, 2040)
  assert dt.DataFrame(s.qBNP).shape == (16, 1)
  assert dt.DataFrame(s.qY).shape == (16, 1)
  assert dt.DataFrame(s.qY[s.s]).shape == (16, 9)
  assert dt.DataFrame(s.qY[s.s], "q").shape == (16, 9)
  assert dt.DataFrame(s.qY[s.s], "s").shape == (16, 9*2)
  s.foo = s.qY.loc[s.sp]
  s.foo.name = "foo"
  assert dt.DataFrame(s.foo).shape == (16, 8)
  with pytest.raises(KeyError):
    dt.DataFrame(s.foo, "q")

def test_DataFrame_with_database():
  dt.REFERENCE_DATABASE = dt.Gdx("../test.gdx")
  s = dt.Gdx("../test.gdx")
  dt.time(2025, 2040)
  assert dt.DataFrame(s, function = lambda s: s.qBNP).shape == (16, 1)
  assert dt.DataFrame(s, function = lambda s: s.qY).shape == (16, 1)
  assert dt.DataFrame(s, function = lambda s: s.qY[s.s]).shape == (16, 9)
  assert dt.DataFrame(s, "q", lambda s: s.qY[s.s]).shape == (16, 9)
  assert dt.DataFrame(s, "q", lambda s: s.pY[s.s] * s.qY[s.s]).shape == (16, 9)
  
def test_DataFrame_with_multiple_baselines():
  b1 = dt.Gdx("../test.gdx")
  b2 = dt.Gdx("../test.gdx")
  s1 = dt.Gdx("../test.gdx")
  s2 = dt.Gdx("../test.gdx")
  s1.reference_database = b1
  s2.reference_database = b2
  b1.qY *= 2
  b2.qY /= 2
  dt.time(2025, 2040)
  assert dt.DataFrame([s1, s2], "q", lambda s: s.pY[s.s] * s.qY[s.s]).shape == (16, 9*2)
  q = dt.DataFrame([s1, s2], "q", lambda s: s.pY['tje'] * s.qY['tje'], ["s1", "s2"])
  assert all(q["s1"] == -0.5)
  assert all(q["s2"] == 1)

  s1 = dt.Gdx("../test.gdx")
  s2 = dt.Gdx("../test.gdx")
  q = dt.DataFrame([s1, s2], "q", lambda s: s.pY['tje'] * s.qY['tje'], ["s1", "s2"], baselines=[b1, b2])
  assert all(q["s1"] == -0.5)
  assert all(q["s2"] == 1)

def test_functions():
  dt.REFERENCE_DATABASE = dt.Gdx("../test.gdx")
  db = dt.Gdx("../test.gdx")
  df1 = dt.DataFrame(db, functions=[lambda s: s.qBNP, lambda s: s.pBNP])
  df2 = dt.DataFrame([db.qBNP, db.pBNP])
  assert df1.equals(df2)

  df1 = dt.DataFrame(db, "pq", functions=[lambda s: s.qBNP, lambda s: s.pBNP])
  df2 = dt.DataFrame([db.qBNP, db.pBNP], "pq")
  assert df1.equals(df2)


"""
Transform Excel work sheet into GAMS code to perform endo exo operations and change variable levels.
"""
import os
import sys
import numpy as np
import pandas as pd
import xlwings as xw
import re

# workbook_path = "C://Users//B031441//Desktop//Konjunktur//Konjunktur//konjunktur.xlsm"
# output_path = "C://Users//B031441//Desktop//Konjunktur//Konjunktur//endo_exo.gms"

def main():
    #  Read execution paremeters
    workbook_path = os.path.abspath(sys.argv[1])
    output_path = os.path.abspath(sys.argv[2])
    convert_workbook(workbook_path, output_path)


def convert_workbook(workbook_path, output_path):
    assert os.path.splitext(workbook_path)[1].lower() in [".xls", ".xlsx", ".xlsm"], \
        f"{workbook_path} is not an Excel workbook"

    # Read Excel file
    print(f"Getting endo exo operations from \"{workbook_path}\"")
    code = ""
    with xw.App(visible=False) as app:
        book = xw.Book(workbook_path)
        endo_exo_range = book.sheets["settings"].range("endo_exo_range").value
        start_year = book.sheets["settings"].range("start_year").value
        for sheet_name in book.sheets["settings"].range("endo_exo_sheets").value:
            if sheet_name is not None:
                print(sheet_name)
                code += convert_sheet(book.sheets[sheet_name], endo_exo_range, start_year)

    with open(output_path, "w") as file:
        print(f"Write endo exo code to {output_path}")
        file.write(code)


def convert_sheet(sheet, endo_exo_range, start_year):
    df = sheet.range(endo_exo_range).options(pd.DataFrame, index=False).value

    # Remove trailing empty rows
    df = df.loc[:df.last_valid_index()]

    # Convert comment column to bool.
    # Row is also counted as column if it contains only 1 non-empty cell
    df["#"] = (df.notna().sum(axis=1) <= 1) | (df["#"].notna())

    # Remove empty columns
    df.dropna(how="all", axis=1, inplace=True)

    header = list(df)

    # First column with values to be used
    year_columns_index = header.index(start_year)

    # Cast years in column header to ints
    header[year_columns_index:] = [int(i) for i in header[year_columns_index:]]
    df.columns = header

    endo_exo_strings = df.apply(
        row_to_gams,
        axis=1,
        result_type="reduce",
        year_columns_index=year_columns_index
    )

    return "\n".join(endo_exo_strings)

def replace_t(text, year):
    """Replace t⨦n] with '{year⨦n}']"""
    def eval_t(m):
        if m[1] is not None:
            return f"'{year + eval(m[1])}']"
        else:
            return f"'{year}']"
    return re.sub(r"t([+-][0-9]+)?\]", eval_t, text)

def row_to_gams(row, year_columns_index):
    """Return GAMS code from row"""
    if row['#']:  # Transfer comments
        return f"# {' | '.join(x for x in row if isinstance(x, str))}"

    code = []
    for year in row.index[year_columns_index:]:
        val = row.loc[year]
        if pd.isnull(val):
            continue

        line = ""
        if row.endo:
            lower = row.endo.replace("[", ".lo[")
            line += replace_t(lower, year) + " = -inf; "
            upper = row.endo.replace("[", ".up[")
            line += replace_t(upper, year) + " = inf; "

        if row.exo:
            assert row.printcode in ["dummy", "p", "n", "m"], f"{row.printcode} is not an implemented print code in row: {row}"
            if row.printcode == "dummy":
                lhs = replace_t(row.exo.replace("[", ".fx["), year)
                rhs = replace_t(row.exo.replace("[", ".l[")), year
                line += f"{lhs} = {rhs};"
            elif row.printcode == "p":
                lhs = replace_t(row.exo.replace("[", ".fx["), year)
                lagged = replace_t(row.exo.replace("[", ".l["), year-1)
                line += f"{lhs} = {lagged} * (1 + ({val}*0.01));"
            elif row.printcode == "m":
                lhs = replace_t(row.exo.replace("[", ".fx["), year)
                lagged = replace_t(row.exo.replace("[", ".l["), year-1)
                line += f"{lhs} = {lagged} + ({val});"
            elif row.printcode == "n":
                lhs = replace_t(row.exo.replace("[", ".fx["), year)
                line += f"{lhs} = {val};"

        code += [line]

    return "\n".join(code)

if __name__ == "__main__":
    main()



from pathlib import Path
import glob
import shutil

def find_root(root_file_name="paths.py", max_iter=20):
    """
    Find the root of the project by looking for a file named root_file_name.
    
    Parameters:
    root_file_name (str): The name of the file to look for (default is "paths.py").
    max_iter (int): Maximum number of parent directories to check before giving up.
    
    Returns:
    str: The path to the directory containing root_file_name, or None if not found.
    """
    current = Path.cwd()
    for _ in range(max_iter):
        if (current / root_file_name).exists():
            return str(current)
        current = current.parent
    return None

def clean_gams_temp_files():
    """
    Clean up temporary files and folders created by GAMS.
    """
    for dir in glob.glob("225*/"):
        shutil.rmtree(dir)


import plotly.io as pio
import pandas
pandas.options.plotting.backend = "plotly"
from .timeseries_analysis import *
from .reporting import *
from .age_plot import *

from .dream_plotly_template import create_dream_template
pio.templates["dream"] = create_dream_template(trace_count=6, items_per_row=2)
pio.templates.default = "dream"

from .dream_plotly_template import dream_colors_rgb, small_figure_layout, large_figure_layout




from .gams_pandas import *

__version__ = "3.3.1"

# Global setting controlling the default name of the time index
X_AXIS_NAME = "t"
# Default axis position if no level named dt.X_AXIS_NAME is found
X_AXIS_INDEX = -1 #  (-1 = last index is time)

# Time settings, used for plot and prt functions
START_YEAR = 1983
END_YEAR = 2060

# Age settings, currently used for plotting only
START_AGE = 0
END_AGE = 100

# Global databases
REFERENCE_DATABASE = None

# Ploting
from .multiindex_plotly import *

# Model specific settings
from .multiindex_plotly.makro_settings import DEFAULT_SET_AGGREGATIONS, AGE_AXIS_TITLE, TIME_AXIS_TITLE, YAXIS_TITLE_FROM_OPERATOR

from .utils import *

from .gamY import gamY


import os
import sys
os.chdir("../..")
sys.path.insert(0, os.getcwd())

import pytest
import numpy as np
import pandas as pd
import dreamtools as dt
from PIL import Image

def test_plotting():
  """
  Plotting is difficult to unit test, but we can at least test that the functions run without errors.
  Visual inspection is required to verify that the plots are correct.
  """
  dt.REFERENCE_DATABASE = dt.Gdx("../test.gdx")
  s = dt.Gdx("../test.gdx")

  dt.time(2025, 20260)
  df1 = dt.DataFrame(
    [s.qC, s.qG, s.qI, s.qX, s.qM],
    names=["Privat forbrug (qC)", "Offentligt forbrug (qG)", "Investeringer (qI)", "Eksport (qX)", "Import (qM)"]
  )
  fig1 = df1.plot()
  assert fig1 == df1.plot()
  
  df2 = dt.DataFrame(
    [s.qC, s.qG, s.qI, s.qX, s.qM],
    "pq",
    names=["Privat forbrug (qC)", "Offentligt forbrug (qG)", "Investeringer (qI)", "Eksport (qX)", "Import (qM)"]
  )
  fig2 = df2.plot(small_figure=True)
  
  df3 = dt.DataFrame(s.qY[s.s], names=list(s.s.texts))
  fig3 = df3.plot()
  assert dt.plot(s.qY[s.s], names=list(s.s.texts)) == fig3

  dt.write_image(fig1, "test1.png", scale=1)
  dt.write_image(fig2, "test2.png", scale=1)
  dt.write_image(fig3, "test3.png", scale=1)

  assert round(Image.open("test1.png").width / 96 * 2.54, 1) == 15.5, "Large figure is not the correct width (pixels × DPI)"
  assert round(Image.open("test2.png").width / 96 * 2.54, 1) == 7.6, "Small figure is not the correct width (pixels × DPI)"
  
def test_yaxis_title():
  dt.REFERENCE_DATABASE = dt.Gdx("../test.gdx")
  s = dt.Gdx("../test.gdx")
  df = dt.DataFrame(s.qY, "m")
  assert df.plot().layout.yaxis.title.text == ""
  assert df.plot(layout=dict(yaxis_title="y-axis title")).layout.yaxis.title.text == ""
  assert df.plot(layout=dict(yaxis_title="y-axis title"), horizontal_yaxis_title=False).layout.yaxis.title.text == "y-axis title"
  


import os
import pandas as pd
import xlwings as xw
from gams import *

# import sys
#  Read any execution paremeters
# args = sys.argv

# assert len(args) > 2, "GAMS path and input path must be specified"
# ws = GamsWorkspace(system_directory=sys.argv[1])
# input_path = os.path.abspath(args[2])
# assert input_path[-4:] == ".gdx", "%s is not a .gdx file" % input_path

ws = GamsWorkspace()
input_path = os.path.abspath("..\\Model\\Gdx\\previous_deep_calibration.gdx")

db = ws.add_database_from_gdx(input_path)

# fp = os.path.join(file_dir, file_name+'_variables.xlsx')
# fp = "../Analysis/Templates/skabelon.xlsx"
fp = "../Konjunktur/konjunktur_ny.xlsx"

# data = pd.read_excel(fp, sheetname=None, na_filter=False)
variables = [symbol for symbol in db if isinstance(symbol, GamsVariable)]
df = pd.DataFrame(
    [(v.name, str(v.domains_as_strings), v.text) for v in variables],
    columns=("Name", "Domains", "Description")
)

wb = xw.Book(fp)
wb.sheets["variables"].range('A1').options(index=False).value = df
wb.save()




import setuptools
import sys
import os
sys.path.insert(0, os.getcwd())
from dreamtools import __version__

with open("README.md", "r") as file:
  long_description = file.read()

setuptools.setup(
  name="dream-tools",
  version=__version__,
  author="Martin Kirk Bonde",
  author_email="martin@bonde.dk",
  description="A collection of tools used by the Danish Research Institute for Economic Analysis and Modelling, DREAM (https://dreamgruppen.dk/).",
  long_description=long_description,
  long_description_content_type="text/markdown",
  url="https://github.com/MartinBonde/dream-tools",
  packages=setuptools.find_packages(),
  classifiers=[
    "Programming Language :: Python :: 3",
    "License :: OSI Approved :: MIT License",
  ],
  python_requires='>=3.6',
  install_requires=["pandas", "numpy", "plotly"],
)


import dreamtools as dt
import numpy as np
import pandas as pd
from inspect import signature
from warnings import simplefilter

simplefilter(action="ignore", category=pd.errors.PerformanceWarning)

def get_reference_database(s=None):
  """Get baseline database associated with a GamsPandasDatabase. Defaults to dt.REFERENCE_DATABASE."""
  if isinstance(s, dt.GamsPandasDatabase) and s.reference_database is not None:
    return s.reference_database
  else:
    return dt.REFERENCE_DATABASE

def time(start, end=None):
  """Set global time settings."""
  if end is None:
    end = start
  dt.START_YEAR = start
  dt.END_YEAR = end

def years():
  """Return list of years in current time settings."""
  return list(range(dt.START_YEAR, dt.END_YEAR+1))

def foo(series):
  """For each level in series, index by ["tot"] if it exists, otherwise index by first element."""
  return series.loc[series.index.get_level_values(series.index.names[0]).isin(["tot"])]

def aggregate_series(series, default_set_aggregations=None):
  """Aggregate series according to default_set_aggregations."""
  if default_set_aggregations is None:
    default_set_aggregations = dt.DEFAULT_SET_AGGREGATIONS

  aggregated = series.copy()
  levels = aggregated.index.levels if aggregated.index.nlevels > 1 else [aggregated.index]
  for level in levels:
    aggregated_level = default_set_aggregations.get(level.name, list(level))
    mask = aggregated.index.get_level_values(level.name).isin(aggregated_level)
    if len(aggregated[mask]) > 0:
      aggregated = aggregated.loc[mask]

  return aggregated

def map_with_baseline(function, data, baselines):
  """Map function to data, passing baselines if function takes two arguments."""
  if len(signature(function).parameters) == 2:
    return list(map(function, data, baselines))
  else:
    return list(map(function, data))

@pd.api.extensions.register_dataframe_accessor("dt")
class _DataFrame(pd.DataFrame):
  """Pandas DataFrame with additional attributes for plotly layout."""

  _internal_names = pd.DataFrame._internal_names + ["layout"]
  _internal_names_set = set(_internal_names)

  @property
  def _constructor(self):
      return _DataFrame

  def plot(
    self,
    operator=None,
    layout={},
    xline=None,
    vertical_legend=True,
    horizontal_yaxis_title=True,
    small_figure=False,
    alternating_dash=None,
    **kwargs
  ):
    """Plot DataFrame using plotly."""
    fig = pd.DataFrame.plot(self, **kwargs)()

    fig.update_layout(**self.layout)

    if vertical_legend:
      fig = dt.vertical_legend(fig)

    if xline is not None:
      fig = add_xline(fig, xline)

    if small_figure:
      fig.update_layout(**dt.small_figure_layout)
    else:
      fig.update_layout(**dt.large_figure_layout)

    fig.update_layout(**layout)

    if horizontal_yaxis_title:
      fig = dt.horizontal_yaxis_title(fig)

    if operator == "s" and alternating_dash is None:
      alternating_dash = "dot"

    if alternating_dash is not None:
      fig = dt.alternating_dash(fig, dash=alternating_dash, line_width=3)

    return fig

def alternating_dash(fig, dash="dot", line_width=None):
  """
  Update traces in plotly figure to have alternating dash styles and reusing colors.
  """
  line_colors = [trace.line.color for trace in fig.data]
  for i, trace in enumerate(fig.data):
    trace.update(
      line_dash = "solid" if i % 2 == 0 else dash,
      line_color = line_colors[i//2],
      line_width = line_width,
    )
  return fig

def horizontal_yaxis_title(fig, text=None):
  """
  Update plotly figure to make the y-axis horizontal using annotations
  """
  if text is None:
    text = fig.layout.yaxis.title.text
  return fig.update_layout(
    yaxis_title_text = "",
    annotations = [
      dict(
        x = 0, xshift = - 0.8 * fig.layout.margin.l, xref = "paper",
        y = 1, yshift = 0.8 * fig.layout.margin.t, yref = "paper",     
        text = text,
        showarrow = False,
      )
    ]
  )

def vertical_legend(fig, col_count=2):
  """
  Update plotly figure by splitting legend into <col_count> columns.
  """
  trace_count = len(fig.data)
  for i, trace in enumerate(fig.data):
    trace.legendgroup = i // (trace_count / col_count)
  return fig

def add_xline(fig, x):
  "Add a vertical line to a plotly figure at x"
  return fig.update_layout(shapes=[dict(
    type='line',
    yref='paper', y0=0, y1=1,
    xref='x', x0=x, x1=x,
    line=dict(
      dash="dash",
    ),
    opacity=0.3,
  )])

def DataFrame(
  data,
  operator=None,
  function=None,
  names=None,
  start_year=None,
  end_year=None,
  baselines=None,
  default_set_aggregations=None,
  functions=None,
):
  if isinstance(data, pd.Series) or isinstance(data, dt.GamsPandasDatabase):
    data = [data]

  if function is not None:
    functions = [function]

  if functions is None:
    if isinstance(data[0], dt.GamsPandasDatabase):
      raise ValueError("Must specify function when passing GamsPandasDatabase.")
    functions = [lambda x: x]

  if baselines is None:
    baselines = [get_reference_database(s) for s in data]

  results = [map_with_baseline(function, data, baselines) for function in functions]

  if operator:
    if None in baselines:
      raise ValueError("Cannot compare with baseline as no baseline has been specified and no global reference database has been set.")
    if isinstance(data[0], pd.Series):
      baseline_series = map(get_baseline_series, data, baselines)
      baseline_results = [map_with_baseline(function, baseline_series, baselines) for function in functions]
    else:
      baseline_results = [map_with_baseline(function, baselines, baselines) for function in functions]
    results = [compare(a, b, operator) for a, b in zip(results, baseline_results)]

  aggregated = [aggregate_series(s, default_set_aggregations) for f in results for s in f]

  df = merge_multiseries(aggregated)
  if start_year is None:
    start_year = dt.START_YEAR
  if end_year is None:
    end_year = dt.END_YEAR
  df = df.loc[start_year:end_year]

  if names:
    df.columns = names

  df = _DataFrame(df)

  # Set default layout for plotly which depends on the operator
  df.layout = {
    "yaxis_title_text": dt.YAXIS_TITLE_FROM_OPERATOR.get(operator, ""),
    "xaxis_title_text": dt.TIME_AXIS_TITLE,
    "legend_title_text": "",
  }

  return df

def get_baseline_series(x, b):
  """Lookup the name of series x in the reference database b and return the series from b with the same index as x"""
  if x.name not in b:
    raise KeyError(f"'{x.name}' was not found in the reference database.")

  y = b[x.name]
  if (
    y.index.nlevels != x.index.nlevels
    or y.index.names != x.index.names
  ):
    raise KeyError(f"The dimension of '{x.name}' is different in the reference database. If indexing a single element write [['element']] rather than ['element'] to prevent the series dimension being reduced.")

  return (x - x + y)[x.index] # Adding and subtracting x is a trick to keep the index

def compare(iter_series, refs, operator):
  """
  Applies an operator to each pair in zip(iter_series, refs)
  """
  if operator in ["q"]:
    return [s / b - 1 for s, b in zip(iter_series, refs)]
  elif operator in ["pq"]:
    return [(s / b - 1)*100 for s, b in zip(iter_series, refs)]
  elif operator in ["m"]:
    return [s - b for s, b in zip(iter_series, refs)]
  elif operator in ["pm"]:
    return [(s - b)*100 for s, b in zip(iter_series, refs)]
  elif operator in ["s"]:
    for b in refs:
      b.name = "baseline." + b.name
    return [i for pair in zip(iter_series, refs) for i in pair]
  elif operator in ["p"]:
    return [s / lag(s) * 100 - 100 for s in iter_series]
  elif operator in ["d"]:
    return [s - lag(s) for s in iter_series]
  elif operator in ["i"]:
    return [index(s) for s in iter_series]
  elif operator in ["log"]:
    return [np.log(s) for s in iter_series]
  elif operator in ["rlog"]:
    return [np.log(s) for s in refs]
  elif operator in ["dlog"]:
    return [np.log(s) - np.log(lag(s)) for s in iter_series]
  else:
    raise ValueError(f"{operator} is not a valid operator.")

def index(series, level=-1, element=None):
  """Return series divided by the value with the index element. E.g. set a series to 1 in a base year."""
  if element is None:
    element = dt.START_YEAR
  if len(series.index.names) > 1:
    return series / series.xs(element, level=level)
  else:
    return series / series[element]

def lag(series, periods=1, lag_axis_index=-1):
  if len(series.index.names) > 1:
    return series.groupby(level=series.index.names[:lag_axis_index:]).shift(periods)
  else:
    return series.shift(periods)

def unstack_multiseries(series, keep_axis_index=-1):
  """
  Return a DataFrame from a Series.
  All levels of the Series are unstacked and concatenated as column names, except <keep_axis_index>.
  """
  if isinstance(series.index, pd.MultiIndex):
    series = series.copy()
    keep_axis_name = series.index.names[keep_axis_index]
    if series.name is None:
      series.name = ""
    series.index = pd.MultiIndex.from_tuples(flatten_keys(series.name, keys, keep_axis_index) for keys in series.index)
    df = series.unstack(0)[series.index.get_level_values(0).unique()]
    df.index.name = keep_axis_name
  else:
    df = pd.DataFrame(series)
  return df

def flatten_keys(name, keys, keep_axis_index):
  keys_str = ','.join(map(str, [keys[i] for i, _ in enumerate(keys) if i != keep_axis_index]))
  flat_name = f"{name}[{keys_str}]"
  return flat_name, keys[keep_axis_index]

def merge_multiseries(series, keep_axis_indices=None):
  """
  Return a DataFrame from any number of Series, with all levels except <keep_axis_index> concatenated as column names.
  """
  output = pd.DataFrame()
  if keep_axis_indices is None:
    keep_axis_indices = [get_keep_axis_index(s) for s in series]
  for s, keep_axis_index in zip(series, keep_axis_indices):
    df = unstack_multiseries(s, keep_axis_index)
    for c in df.columns:
      new_name = c if (c != "0") else ""
      iter = 0
      while new_name in output:
        iter += 1
        new_name = f"{c}{iter}"
      output[new_name] = df[c]

  return output


def get_keep_axis_index(series):
  """
  Return the index of the axis to keep when unstacking a multi-indexed series.
  """
  try:
    return series.index.names.index(dt.X_AXIS_NAME)
  except ValueError:
    return dt.X_AXIS_INDEX # Default if no level named dt.X_AXIS_NAME is found



from math import ceil
from time import sleep
import inspect

from plotly import offline as pyo
from plotly.subplots import make_subplots

import pandas as pd

import dreamtools as dt

def plot(
    data,
    operator=None,
    *args,
    **kwargs
  ):
  """Shorthand for DataFrame(...).plot(...)"""
  dataframe_kwargs = {k: v for k, v in kwargs.items() if k in inspect.signature(dt.DataFrame).parameters}
  plot_kwargs = {k: v for k, v in kwargs.items() if k not in dataframe_kwargs}
  dataframe = dt.DataFrame(data, operator, *args, **dataframe_kwargs)
  return dataframe.plot(operator, **plot_kwargs)

def prt(
  iter_series,
  operator=None,
  function=None,
  names=None,
  start_year=None,
  end_year=None,
  reference_database=None,
  default_set_aggregations=None,
  dec=6,
  max_rows=100,
  max_columns=20,
):
  """Print a table of a series or list of series."""
  df = dt.DataFrame(iter_series, operator, function, names, start_year, end_year, reference_database, default_set_aggregations)
  df.style.set_properties(**{"text-align": "right", "precision": dec})
  with pd.option_context('display.max_rows', max_rows, 'display.max_columns', max_columns):  # more options can be specified also
    display(df)

def write_image(fig, file_name, scale=3):
  fig.write_image(file_name, scale=scale)
  if file_name.endswith(".png"):
    from PIL import Image
    with Image.open(file_name) as img:
      sleep(0.01)
      img.save(file_name, dpi=(96 * scale, 96 * scale))

def figures_to_html(figs, filename="figures.html", encoding="utf-8"):
  """Write an iter of plotly figures to an html file."""
  with open(filename, 'w', encoding=encoding) as file:
    file.write("<html><head></head><body>" + "\n")

    for i, fig in enumerate(figs):
      inner_html = pyo.plot(
        fig, include_plotlyjs=(i==0), output_type='div'
      )
      file.write(inner_html)

    file.write("</body></html>" + "\n")

def subplot(figures, n_columns=2, n_legend_columns=2, **kwargs):
  """Create subplot from iter of figures."""
  rows = ceil(len(figures) / n_columns)
  fig = make_subplots(
    rows=rows,
    cols=n_columns,
    subplot_titles=[f["layout"]["title"]["text"] for f in figures],
    **kwargs
  )
  for i, f in enumerate(figures):
    for trace in f["data"]:
      trace.showlegend = i==1
      fig.add_trace(trace, row=ceil((i+1)/n_columns), col=1 + i % n_columns)
  trace_count = len(figures[0].data)
  for i, trace in enumerate(fig.data):
    trace.legendgroup = i // (trace_count / n_legend_columns)
  return fig

def html_table(df, precision=2, header=True, header_background_color="#14AFA6", head_text_color="#ffffff"):
  """Return string with a formated HTML version of a Pandas Dataframe"""
  df = df.copy()
  df.index.name, df.columns.name = None, None

  styles = [
    # table properties
    dict(selector=" ",
         props=[
           ("margin", "0"),
           ("font-family", '"Hind", "Helvetica", "Arial", sans-serif'),
           ("border-collapse", "collapse"),
           # ("border", "2px solid #0F837D"),
           ("text-align", "right"),
         ]),

    # header color
    dict(selector="thead",
         props=[
           ("background-color", header_background_color),
           ("color", head_text_color),
         ]),

    # background shading
    dict(selector="tbody",
         props=[("background-color", "#E6E6E8")]),
    # dict(selector="tbody tr:nth-child(even)",
    #         props=[("background-color", "#fff")]),
    # dict(selector="tbody tr:nth-child(odd)",
    #         props=[("background-color", "#eee")]),

    # body cell properties
    dict(selector="td",
         props=[
           ("font-size", "120%"),
           ("padding", ".5em"),
           ("border", "1px solid #ffffff"),
         ]),

    # header cell properties
    dict(selector="th",
         props=[
           ("font-size", "120%"),
           ("text-align", "right"),
           ("padding", ".5em"),
           ("border", "1px solid #ffffff"),
         ]),

    # caption placement
    dict(selector="caption",
         props=[("caption-side", "bottom")]),
  ]
  if not header:
    styles.append({"selector": "thead", "props": [("font-size", "0pt")]})

  with pd.option_context("precision", precision):
    return "<meta charset='UTF-8'>" + df.style.set_table_styles(styles).render()